@@ -2,7 +2,7 @@
 # ==============================================================================
 # portsreinstall library script
 # - Wrappers for hiding version differences in the Ports/Packages system -
-# Copyright (C) 2013-2014 Mamoru Sakaue, MwGhennndo, All Rights Reserved.
+# Copyright (C) 2013-2015 Mamoru Sakaue, MwGhennndo, All Rights Reserved.
 # This software is distributed under the 2-Clause BSD License.
 # ==============================================================================
 
@@ -48,16 +48,49 @@ pkgsys_pkgchksum_file ()
 	echo CHECKSUM.MD5
 }
 
+# ============= Get the origin of the currently used pkg(8) =============
+pkgsys_portsmgmt_pkg ()
+{
+	local origin
+	if [ ! -e "${DBDIR}/PKG_ORIGIN" ]
+	then
+		origin=`pkgsys_sysvar PKG_ORIGIN` || :
+		[ -n "$origin" -a -d "${PORTSDIR}/$origin" ] || origin=ports-mgmt/pkg
+		echo "$origin" > ${DBDIR}/PKG_ORIGIN
+	fi
+	cat "${DBDIR}/PKG_ORIGIN"
+}
+
+# ============= Get the origin of the currently used dialog4ports(1) =============
+pkgsys_portsmgmt_dialog4ports ()
+{
+	echo "$origin"
+	local origin
+	if [ ! -e "${DBDIR}/DIALOGPORT" ]
+	then
+		origin=`pkgsys_sysvar DIALOGPORT` || :
+		[ -n "$origin" -a -d "${PORTSDIR}/$origin" ] || origin=ports-mgmt/dialog4ports
+		echo "$origin" > ${DBDIR}/DIALOGPORT
+	fi
+	cat "${DBDIR}/DIALOGPORT"
+}
+
 # ============= Check whether a port is indispensable for the standard function of the ports/packages system =============
 pkgsys_is_pkgtool ()
 {
 	case $1 in
-	ports-mgmt/pkg|ports-mgmt/dialog4ports)
+	ports-mgmt/pkg|ports-mgmt/pkg-devel|ports-mgmt/dialog4ports)
 		;;
 	*)	return 1;;
 	esac
 }
 
+# ============= Check whether a port is indispensable for package operations =============
+pkgsys_is_necessary_pkgtool ()
+{
+	[ x"$WITH_PKGNG" = x'yes' -a \( x"$1" = x'ports-mgmt/pkg' -o x"$1" = x'ports-mgmt/pkg-devel' \) ]
+}
+
 # ============= Check whether the dialog for selecting port options is dialog4ports =============
 pkgsys_is_dialog4ports_used ()
 {
@@ -273,7 +306,7 @@ pkgsys_def_pkgtools ()
 			then
 				pkg -N 2> /dev/null && return
 				env ASSUME_ALWAYS_YES=yes pkg bootstrap -f
-				pkg_is_tool_available
+				pkg -N 2> /dev/null
 			else
 				which -s pkg && return
 			fi
@@ -364,6 +397,7 @@ pkgsys_def_pkgtools ()
 		pkg_delete_f ()
 		{
 			pkg delete -fqy "$@"
+			pkg -N 2> /dev/null || return 0	# If pkg(8) is deinstalled successfully
 			pkg info -e "$@" || return 0	# Countermeasure for a bug found for pkg-1.3.4 (at least not until 1.2.7_4)
 			pkg delete -fy "$@"
 		}
@@ -489,7 +523,8 @@ pkgsys_def_pkgtools ()
 			PKGNG_AUTODEPS=NO
 			PKGNG_PORTAUDIT_SITE='http=//portaudit.FreeBSD.org/auditfile.tbz'
 			# Load configuration for pkg(1)
-			pkg_conf=`pkg query %Fp pkg | grep '/etc/pkg\.conf\.sample$' | sed 's/\.sample$//'`
+			pkg_conf=`pkg query %Fp pkg 2> /dev/null | grep '/etc/pkg\.conf\.sample$' | sed 's/\.sample$//'` || :
+			[ -n "$pkg_conf" ] || pkg_conf=${MYPREFIX}/etc/pkg.conf
 			grep -v -e '^[[:space:]]*#' -e '^[[:space:]]*$' "$pkg_conf" 2> /dev/null \
 				| grep -e '^[[:space:]]*[A-Z0-9_]*[[:space:]]*:[[:space:]]*.*' \
 				| while read srcline
@@ -504,7 +539,7 @@ pkgsys_def_pkgtools ()
 		pkg_rescue_tools ()
 		{
 			local packagepath checksumpath pkgname is_successful
-			packagepath=`pkgsys_get_backup_pkg ports-mgmt/pkg` && \
+			packagepath=`pkgsys_get_backup_pkg \`pkgsys_portsmgmt_pkg\`` && \
 				pkg_add_tools "$packagepath" && return
 			pkg_is_tool_available && return
 			message_echo "WARNING: WITH_PKGNG is set, but pkgng is still missing. It is installed now." >&2
@@ -521,10 +556,10 @@ pkgsys_def_pkgtools ()
 				unset WITH_PKGNG
 				unset WITHOUT_PKGNG
 				
-				message_echo "INFO: Attempting deinstallation of ports-mgmt/pkg to make sure."
-				env __MAKE_CONF="${TMPDIR}/make.conf" make -C "${PORTSDIR}/ports-mgmt/pkg" deinstall || :
-				message_echo "INFO: Attempting (re)installation by ports-mgmt/pkg."
-				env __MAKE_CONF="${TMPDIR}/make.conf" make -C "${PORTSDIR}/ports-mgmt/pkg" reinstall clean
+				message_echo "INFO: Attempting deinstallation of pkg(8) to make sure."
+				env __MAKE_CONF="${TMPDIR}/make.conf" make -C "${PORTSDIR}/`pkgsys_portsmgmt_pkg`" deinstall || :
+				message_echo "INFO: Attempting (re)installation by pkg(8)."
+				env __MAKE_CONF="${TMPDIR}/make.conf" make -C "${PORTSDIR}/`pkgsys_portsmgmt_pkg`" reinstall clean
 			) && {
 				pkg2ng || :
 				pkg_is_tool_available
@@ -828,7 +863,7 @@ pkgsys_eval_ports_glob ()
 # ============= Create a back-up package archive =============
 pkgsys_create_backup_pkg ()
 {
-	local pkgname dstdir origin backup_pkg_old origin_regexp pkgname_ptn backup_pkg pkgpath
+	local pkgname dstdir origin backup_pkg_old pkgname_ptn backup_pkg dbpath pkgpath origin_regexp
 	pkgname=$1
 	dstdir=$2
 	rm -rf "${TMPDIR}"/package.tmp
@@ -851,6 +886,9 @@ pkgsys_create_backup_pkg ()
 		message_echo "WARNING: Failed to create backup package for $pkgname." >&2
 		return 1
 	fi
+	dbpath=${DBDIR}/backup/$origin
+	[ -d "$dbpath" ] || mkdir -p "$dbpath"
+	pkg_info_qL > $dbpath/previously_installed_files
 	[ -d "$dstdir" ] || mkdir -p "$dstdir"
 	mv "${TMPDIR}/package.tmp/$backup_pkg" "$dstdir"
 	pkgpath=$dstdir/$backup_pkg
@@ -890,7 +928,7 @@ pkgsys_delete_backup_pkg ()
 # ============= Get an existing package archive path for a port origin =============
 pkgsys_get_backup_pkg ()
 {
-	local origin origin_regexp
+	local origin tmpnewest origin_regexp
 	origin=$1
 	tmpnewest=${TMPDIR}/pkgsys_get_backup_pkg::newest
 	origin_regexp=`str_escape_regexp "$origin"`
@@ -906,6 +944,27 @@ pkgsys_get_backup_pkg ()
 	cat "$tmpnewest" 2> /dev/null
 }
 
+# ============= Get a file list to be restored by the current backup package for a port origin =============
+pkgsys_get_restored_files_by_backup_pkg ()
+{
+	local origin
+	origin=$1
+	cat "${DBDIR}/backup/$origin/previously_installed_files" 2> /dev/null || :
+}
+
+# ============= Check whether any file match restored files by the current backup package for a port origin =============
+pkgsys_chk_match_to_restored_files_by_backup_pkg ()
+{
+	local origin
+	origin=$1
+	pattern_file=$2
+	while read pattern
+	do 
+		grep -qE "$pattern" "${DBDIR}/backup/$origin/previously_installed_files" && return
+	done < $pattern_file
+	false
+}
+
 # ============= Get a package name from a package archive file name =============
 pkgsys_pkgarc_to_pkgname ()
 {
@@ -1013,7 +1072,10 @@ pkgsys_get_conflicting_installed_pkgs ()
 	mode=$1
 	origin=$2
 	tmp_conflicts=${TMPDIR}/pkgsys_get_conflicting_installed_pkgs::conflicts
-	pkg_info_Eg `pkgsys_get_conflicting_pkgs_patterns "$mode" "$origin"` > $tmp_conflicts || :
+	pkgsys_get_conflicting_pkgs_patterns "$mode" "$origin" | while read pkg_pattern
+	do
+		pkg_info_Eg "$pkg_pattern" || :
+	done > $tmp_conflicts
 	cat "$tmp_conflicts"
 	[ `wc -l < $tmp_conflicts` -gt 0 ]
 }	
