@@ -2,7 +2,7 @@
 # ==============================================================================
 # portsreinstall library script
 # - Reinstallation processes -
-# Copyright (C) 2013-2014 Mamoru Sakaue, MwGhennndo, All Rights Reserved.
+# Copyright (C) 2013-2015 Mamoru Sakaue, MwGhennndo, All Rights Reserved.
 # This software is distributed under the 2-Clause BSD License.
 # ==============================================================================
 
@@ -28,10 +28,19 @@ reinstall_skip ()
 	fileedit_rm_a_line "$REINSTALL_ORIGIN" "${DBDIR}/stage.loop_list/reinst_todo.remain"
 }
 
+# ============= Check whether a package is forbidden due to conflicts =============
+reinstall_quick_chk_forbidden_conflicts ()
+{
+	local pkg pkg_regexp_esc
+	pkg=$1
+	pkg_regexp_esc=`str_escape_regexp "$pkg"`
+	grep -qE "^$pkg_regexp_esc:" "${DBDIR}/forbidden_conflicts"
+}
+
 # ============= Check whether a package is forbidden due to conflicts =============
 reinstall_chk_forbidden_conflicts ()
 {
-	local pkg tmp_forbidden backup_pkg pkg pkg_regexp_esc
+	local pkg tmp_forbidden pkg_regexp_esc
 	pkg=$1
 	tmp_forbidden=${TMPDIR}/reinstall_chk_forbidden_conflicts:forbidden
 	pkg_regexp_esc=`str_escape_regexp "$pkg"`
@@ -229,6 +238,35 @@ reinstall_backup_and_delete_conflicts ()
 	mv "${DBDIR}/deleted_conflicts.tmp" "${DBDIR}/deleted_conflicts"
 }
 
+# ============= Back up and delete remaining actual install conflicts =============
+reinstall_backup_and_delete_remaining_install_conflicts ()
+{
+	local stagedir tmp_conflicts
+	tmp_conflicts=${TMPDIR}/reinstall_backup_and_delete_remaining_install_conflicts
+	message_echo "(Checking installation conflict...)"
+	stagedir=`database_query_get_makevar_val "$REINSTALL_ORIGIN" STAGEDIR`
+	{
+		( set -e
+			cd "$stagedir"
+			find . -not -type d
+		) | sed 's|^\.||' | while read filepath
+		do
+			[ ! -e "$filepath" ] && continue
+			pkg=`pkg_which "$filepath" || :`
+			[ -z "$pkg" ] && continue
+			origin=`pkg_info_qo "$pkg" || :`
+			[ x"$origin" = x"$REINSTALL_ORIGIN" ] || echo "$pkg"
+		done
+		pkgsys_get_conflicting_installed_pkgs install "$REINSTALL_ORIGIN"
+	} | sort -u > $tmp_conflicts
+	while read pkg_conflict
+	do
+		fileedit_add_a_line_if_new "$pkg_conflict:$REINSTALL_NEWPKGNAME" \
+			"${DBDIR}/forbidden_conflicts"
+	done < $tmp_conflicts
+	reinstall_backup_and_delete_conflicts < $tmp_conflicts
+}
+
 # ============= Restoration of backed up conflicts =============
 reinstall_restore_conflicts ()
 {
@@ -403,13 +441,21 @@ reinstall_deinstall ()
 {
 	[ -n "$REINSTALL_CURRENTPKG" ] || return 0
 	reinstall_chk_stage_complete DEINSTALL && return
-	echo "$REINSTALL_CURRENTPKG" | tr ' ' '\n' | grep -v '^$' | while read pkg
-	do
-		pkg_delete_f "$pkg" || \
+	if pkgsys_is_necessary_pkgtool "$REINSTALL_ORIGIN"
+	then
+		pkg_delete_f "$REINSTALL_CURRENTPKG" || \
 		{
-			message_echo "WARNING: Failed to deinstall $pkg by $PKGSYS_CMD_PKG_DELETE." >&2
+			message_echo "WARNING: Failed to deinstall $REINSTALL_CURRENTPKG by $PKGSYS_CMD_PKG_DELETE." >&2
 		}
-	done
+	else
+		echo "$REINSTALL_CURRENTPKG" | tr ' ' '\n' | grep -v '^$' | while read pkg
+		do
+			pkg_delete_f "$pkg" || \
+			{
+				message_echo "WARNING: Failed to deinstall $pkg by $PKGSYS_CMD_PKG_DELETE." >&2
+			}
+		done
+	fi
 	message_echo "-- (Trying to deinstall by ports to make sure. This usually ends up with warnings.)"
 	reinstall_make deinstall || \
 	{
@@ -642,7 +688,8 @@ reinstall_exec ()
 		then
 			if ! reinstall_chk_stage FAILED_INSTALL_BY_PKG
 			then
-				if [ "x$REINSTALL_ORIGIN" = 'xports-mgmt/pkg'  -a "x$PKGSYS_USE_PKGNG" = xyes ]
+				if [ "x$PKGSYS_USE_PKGNG" = xyes ] \
+					&& pkgsys_is_necessary_pkgtool "$REINSTALL_ORIGIN"
 				then
 					if reinstall_deinstall && pkg_is_tool_available && pkg_info_e "$REINSTALL_NEWPKGNAME"
 					then
@@ -671,6 +718,10 @@ reinstall_exec ()
 						else
 							reinstall_register_stage FAILED_INSTALL_BY_PKG
 						fi
+						if pkgsys_is_necessary_pkgtool "$REINSTALL_ORIGIN"
+						then
+							pkg_is_tool_available || pkg_rescue_tools
+						fi
 					else
 						reinstall_register_stage FAILED_INSTALL_BY_PKG
 					fi
@@ -816,13 +867,6 @@ reinstall_exec ()
 		then
 			[ $flag_restarted_build = yes ] && message_echo "INFO: This failure may be due to restarting from a terminated build."
 			is_build_env_modified=no
-			if pkg_info_eO "$REINSTALL_ORIGIN"
-			then
-				message_echo "INFO: The currently installed package for this port is deinstalled for retrial because it may have effects on build."
-				reinstall_pkg_backup || :
-				reinstall_deinstall
-				is_build_env_modified=yes
-			fi
 			if pkgsys_get_conflicting_installed_pkgs install "$REINSTALL_ORIGIN" \
 				> ${TMPDIR}/reinstall_exec::conflicts_install
 			then
@@ -830,8 +874,19 @@ reinstall_exec ()
 				reinstall_backup_and_delete_conflicts < ${TMPDIR}/reinstall_exec::conflicts_install
 				is_build_env_modified=yes
 			fi
+			if pkg_info_eO "$REINSTALL_ORIGIN"
+			then
+				message_echo "INFO: The currently installed package for this port is deinstalled for retrial because it may have effects on build."
+				reinstall_pkg_backup || :
+				reinstall_deinstall
+				is_build_env_modified=yes
+			fi
 			if [ $flag_restarted_build = no -a $is_build_env_modified = no ]
 			then
+				if pkgsys_is_necessary_pkgtool "$REINSTALL_ORIGIN"
+				then
+					pkg_is_tool_available || pkg_rescue_tools
+				fi
 				reinstall_restore_conflicts "$REINSTALL_ORIGIN"
 				reinstall_restore_if_temporarily_deinstalled
 				reinstall_errproc 'build'
@@ -842,6 +897,10 @@ reinstall_exec ()
 			message_echo "INFO: Retrying the build process after cleaning for $REINSTALL_ORIGPKGTAG."
 			reinstall_make clean || \
 			{
+				if pkgsys_is_necessary_pkgtool "$REINSTALL_ORIGIN"
+				then
+					pkg_is_tool_available || pkg_rescue_tools
+				fi
 				reinstall_restore_conflicts "$REINSTALL_ORIGIN"
 				reinstall_restore_if_temporarily_deinstalled
 				reinstall_errproc 'build and clean after build failure'
@@ -851,6 +910,10 @@ reinstall_exec ()
 			reinstall_register_stage_complete CLEAN_BEFORE_BUILD
 			reinstall_make build MAKE_JOBS_UNSAFE=yes || \
 			{
+				if pkgsys_is_necessary_pkgtool "$REINSTALL_ORIGIN"
+				then
+					pkg_is_tool_available || pkg_rescue_tools
+				fi
 				reinstall_restore_conflicts "$REINSTALL_ORIGIN"
 				reinstall_restore_if_temporarily_deinstalled
 				reinstall_errproc 'retrial of build after failure'
@@ -858,16 +921,18 @@ reinstall_exec ()
 				return
 			}
 		fi
+		if pkgsys_is_necessary_pkgtool "$REINSTALL_ORIGIN"
+		then
+			pkg_is_tool_available || pkg_rescue_tools
+		fi
 		reinstall_register_stage_complete BUILD
 	fi
 	reinstall_deregister_stage in_build
 	if reinstall_chk_stage in_stage
 	then
 		message_echo "(Restarting the previously terminated staging process...)"
-		flag_restarted_stage=yes
 	else
 		reinstall_register_stage in_stage
-		flag_restarted_stage=no
 	fi
 	if ! reinstall_chk_stage_complete STAGE
 	then
@@ -885,10 +950,8 @@ reinstall_exec ()
 	if reinstall_chk_stage in_install
 	then
 		message_echo "(Restarting the previously terminated installation process...)"
-		flag_restarted_build=yes
 	else
 		reinstall_register_stage in_install
-		flag_restarted_build=no
 	fi
 	reinstall_pkg_backup || :
 	if [ -e "$REINSTALL_DBNODE_DIR/BEFOREDEINSTALL.conf" ] && ! reinstall_chk_stage_complete BEFOREDEINSTALL
@@ -908,22 +971,16 @@ reinstall_exec ()
 		fi
 		reinstall_register_stage_complete BEFOREDEINSTALL
 	fi
-	reinstall_deinstall
 	if ! reinstall_chk_stage_complete INSTALL
 	then
-		message_echo "(Checking installation conflict...)"
-		stagedir=`database_query_get_makevar_val "$REINSTALL_ORIGIN" STAGEDIR`
-		{
-			( set -e
-				cd "$stagedir"
-				find . -not -type d
-			) | sed 's|^\.||' | while read filepath
-			do
-				[ -e "$filepath" ] || continue
-				pkg_which "$filepath"
-			done
-			pkgsys_get_conflicting_installed_pkgs install "$REINSTALL_ORIGIN"
-		} | sort -u | reinstall_backup_and_delete_conflicts
+		if pkgsys_is_necessary_pkgtool "$REINSTALL_ORIGIN"
+		then
+			reinstall_backup_and_delete_remaining_install_conflicts
+			reinstall_deinstall
+		else
+			reinstall_deinstall
+			reinstall_backup_and_delete_remaining_install_conflicts
+		fi
 		if ! reinstall_chk_stage FAILED_INSTALL
 		then
 			if reinstall_make $insttarget || \
@@ -943,7 +1000,12 @@ reinstall_exec ()
 		fi
 		if reinstall_chk_stage FAILED_INSTALL
 		then
-			reinstall_failed_install_recover
+			if pkgsys_is_necessary_pkgtool "$REINSTALL_ORIGIN"
+			then
+				pkg_is_tool_available || pkg_rescue_tools
+			else
+				reinstall_failed_install_recover
+			fi
 			if [ -e "$REINSTALL_DBNODE_DIR/AFTERINSTALL.conf" ] && \
 				! reinstall_chk_stage_complete FAILED_INSTALL.AFTERINSTALL
 			then
@@ -967,6 +1029,29 @@ reinstall_exec ()
 		fi
 	fi
 	reinstall_restore_if_temporarily_deinstalled
+	if ! reinstall_chk_stage_complete UPDATE_CONFLICTS
+	then
+		pkg_info_qL "$REINSTALL_CURRENTPKG" | str_escape_regexp_filter | sed 's/^/^/; s/$/$/' \
+			> ${TMPDIR}/previously_installed_files.grep_pattern
+		cat "${DBDIR}/backup_pkgarcs.lst" 2> /dev/null \
+			| while read origin_bak pkgpath_bak
+		do
+			pkg_bak=`pkgsys_pkgarc_to_pkgname "$pkgpath_bak"`
+			if [ "$origin_bak" = "$REINSTALL_ORIGIN" ]
+			then
+				[ "$pkg_bak" = "$REINSTALL_NEWPKGNAME" ] && continue
+			elif reinstall_quick_chk_forbidden_conflicts "$pkg_bak"
+			then
+				continue
+			elif ! pkgsys_chk_match_to_restored_files_by_backup_pkg "$origin_bak" "${TMPDIR}/previously_installed_files.grep_pattern"
+			then
+				continue
+			fi
+			fileedit_add_a_line_if_new "$pkg_bak:$REINSTALL_NEWPKGNAME" \
+				"${DBDIR}/forbidden_conflicts"
+		done
+		reinstall_register_stage_complete UPDATE_CONFLICTS
+	fi
 	if [ -e "$REINSTALL_DBNODE_DIR/AFTERINSTALL.conf" ] && ! reinstall_chk_stage_complete AFTERINSTALL
 	then
 		message_echo "-- AFTERINSTALL operations (start)"
